from .base import BaseTestCase
from subconscious.column import Integer, Column
from subconscious.model import RedisModel, BadDataError


class TestUser(RedisModel):
    id = Integer(primary_key=True, auto_increment=True)
    name = Column(type=str)
    auto_id = Integer(auto_increment=True)


class TestAutoInt(BaseTestCase):

    def test_auto(self):
        user = TestUser(name='foo')
        self.loop.run_until_complete(user.save(self.db))
        self.assertEqual(user.id, 1)
        user = TestUser()
        self.loop.run_until_complete(user.save(self.db))
        self.assertEqual(user.id, 2)

    def test_setting_auto_increment_should_throw_bad_data_error(self):
        # via constructor
        with self.assertRaises(BadDataError):
            TestUser(id=777)

        user = TestUser()
        # via attribute mutation
        with self.assertRaises(BadDataError):
            user.id = 424

    def test_load(self):
        user = TestUser(name='foo')
        self.loop.run_until_complete(user.save(self.db))
        user_in_db = self.loop.run_until_complete(TestUser.load(db=self.db, identifier=1))
        self.assertEqual(user.name, user_in_db.name)

    def test_update(self):
        user = TestUser(name='foo')
        self.loop.run_until_complete(user.save(self.db))
        user_in_db = self.loop.run_until_complete(TestUser.load(db=self.db, identifier=1))
        self.assertEqual(user.name, user_in_db.name)
        user_in_db.name = 'bar'
        self.loop.run_until_complete(user_in_db.save(self.db))
        user_in_db = self.loop.run_until_complete(TestUser.load(db=self.db, identifier=1))
        self.assertEqual('bar', user_in_db.name)



from subconscious.model import RedisModel, Column, InvalidQuery
from uuid import uuid1
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'


class TestUser(RedisModel):
    id = Column(primary_key=True)
    name = Column(index=True, sort=True)
    age = Column(index=True, type=int,)
    locale = Column(index=True, type=int, required=False)
    status = Column(type=str, enum=StatusEnum)


class TestAll(BaseTestCase):

    def setUp(self):
        super(TestAll, self).setUp()
        user_id = str(uuid1())
        user = TestUser(id=user_id, name='Test name', age=100, status='active')
        ret = self.loop.run_until_complete(user.save(self.db))
        self.assertTrue(ret)

        user_id = str(uuid1())
        user1 = TestUser(id=user_id, name='ZTest name', age=53)
        ret = self.loop.run_until_complete(user1.save(self.db))
        self.assertTrue(ret)

        user_id = str(uuid1())
        user1 = TestUser(id=user_id, name='Test name2', age=53)
        ret = self.loop.run_until_complete(user1.save(self.db))
        self.assertTrue(ret)

    def test_all(self):
        async def _test_all():
            async for x in TestUser.all(db=self.db):
                self.assertEqual(type(x), TestUser)
                self.assertTrue(x.name in ('Test name', 'ZTest name', 'Test name2'))
                self.assertTrue(x.age in (100, 53))
        self.loop.run_until_complete(_test_all())

    def test_all_with_order(self):
        async def _test():

            expected_in_order = ['Test name', 'Test name2', 'ZTest name']
            result_list = []
            async for x in TestUser.all(db=self.db, order_by='name'):
                result_list.append(x.name)
            self.assertEqual(result_list, expected_in_order)

            expected_in_order.sort(reverse=True)
            result_list = []
            async for x in TestUser.all(db=self.db, order_by='-name'):
                result_list.append(x)
            self.assertEqual([x.name for x in result_list], expected_in_order)

            # update a record to force sort order change
            result_list[0].name = 'AATest name'
            await result_list[0].save(self.db)
            result_list = []
            expected_in_order = ['AATest name', 'Test name', 'Test name2']
            async for x in TestUser.all(db=self.db, order_by='name'):
                result_list.append(x)
            self.assertEqual([x.name for x in result_list], expected_in_order)

        self.loop.run_until_complete(_test())

    def test_filter_by_non_existing_fields_should_fail(self):
        async def _test():
            async for x in TestUser.filter_by(db=self.db, non_existing1='dummy', non_existing2=1):
                assert x  # Just to satisfy flake8
        with self.assertRaises(InvalidQuery):
            self.loop.run_until_complete(_test())

    def test_filter_by_non_indexed_field_should_fail(self):
        async def _test():
            async for x in TestUser.filter_by(db=self.db, status='active',):
                assert x  # Just to satisfy flake8
        with self.assertRaises(InvalidQuery):
            self.loop.run_until_complete(_test())

    def test_all_iter(self):
        names_in_expected_order = ['Test name', 'Test name2', 'ZTest name']
        result_array = []

        async def _test_loop():
            count = 0
            async for x in TestUser.all(db=self.db, order_by='name'):
                self.assertEqual(x.name, names_in_expected_order[count])
                count += 1
                result_array.append(x.name)
            self.assertEqual(names_in_expected_order, result_array)

        self.loop.run_until_complete(_test_loop())


class TestAllLimitOffset(TestAll):

    def test_limit_only(self):
        async def _test():
            result_array = []
            async for x in TestUser.all(db=self.db, order_by='name', limit=1):
                result_array.append(x.name)
            self.assertEqual(result_array, ['Test name'])
        self.loop.run_until_complete(_test())

    def test_limit_and_offset(self):
        async def _test():
            result_array = []
            async for x in TestUser.all(db=self.db, order_by='name', limit=1, offset=1):
                result_array.append(x.name)
            self.assertEqual(result_array, ['Test name2'])
        self.loop.run_until_complete(_test())

    def test_offset_only(self):
        async def _test():
            result_array = []
            async for x in TestUser.all(db=self.db, order_by='name', offset=1):
                result_array.append(x.name)
            self.assertEqual(result_array, ['Test name2', 'ZTest name'])
        self.loop.run_until_complete(_test())

    def test_over_offset(self):
        async def _test():
            result_array = []
            async for x in TestUser.all(db=self.db, order_by='name', offset=999):
                result_array.append(x.name)
            self.assertEqual(result_array, [])
        self.loop.run_until_complete(_test())

    def test_nonbinding_limit(self):
        async def _test():
            result_array = []
            async for x in TestUser.all(db=self.db, order_by='name', limit=999):
                result_array.append(x.name)
            self.assertEqual(result_array, ['Test name', 'Test name2', 'ZTest name'])
        self.loop.run_until_complete(_test())



class Query(object):
    def __init__(self, model, db):
        self._model = model
        self._filter = {}
        self._order_by = None
        self._limit = None
        self._offset = None
        self._db = db

    def filter(self, **kwargs):
        self._filter.update(kwargs)
        return self

    def order_by(self, order_by):
        self._order_by = order_by
        return self

    def limit(self, limit):
        self._limit = limit
        return self

    def offset(self, offset):
        self._offset = offset
        return self

    def __aiter__(self):
        self.result_set = self._model.filter_by(
            db=self._db,
            order_by=self._order_by,
            limit=self._limit,
            offset=self._offset,
            **self._filter,)

        return self

    async def __anext__(self):
        async for x in self.result_set:
            return x
        raise StopAsyncIteration

    async def first(self):
        return await self._model.get_object_or_none(db=self._db, order_by=self._order_by, **self._filter)



from subconscious.model import RedisModel, Column
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'


class Diner(RedisModel):
    table_num = Column(composite_key=True, type=int)
    seat_num = Column(composite_key=True, type=int)
    comments = Column(type=str)


class TestCompositeKeys(BaseTestCase):

    def setUp(self):
        super(TestCompositeKeys, self).setUp()

        diner = Diner(
            table_num=1,
            seat_num=4,
            comments='Very polite',
        )
        ret = self.loop.run_until_complete(diner.save(self.db))
        self.assertTrue(ret)

    def test_valid_composite_key_should_return(self):
        async def _test():
            count = 0
            async for x in Diner.filter_by(db=self.db, table_num=1, seat_num=4,):
                self.assertEqual(x.comments, 'Very polite')
                count += 1
            self.assertEqual(count, 1)

        self.loop.run_until_complete(_test())

    def test_partial_composite_key_should_succeed(self):
        # FIXME: is this really the desired behavior?
        async def _test():
            count = 0
            async for x in Diner.filter_by(db=self.db, table_num=1,):
                self.assertEqual(x.comments, 'Very polite')
                count += 1
            self.assertEqual(count, 1)
        self.loop.run_until_complete(_test())



from subconscious.model import RedisModel, Column
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'


class Diner(RedisModel):
    table_num = Column(composite_key=True, type=int)
    seat_num = Column(composite_key=True, type=int)
    comments = Column(type=str)


class TestGetObectOrNone(BaseTestCase):
    def setUp(self):
        super(TestGetObectOrNone, self).setUp()

        diner = Diner(
            table_num=1,
            seat_num=4,
            comments='Very polite',
        )
        ret = self.loop.run_until_complete(diner.save(self.db))
        self.assertTrue(ret)
        diner = Diner(
            table_num=2,
            seat_num=5,
            comments='Very rude',
        )
        ret = self.loop.run_until_complete(diner.save(self.db))
        self.assertTrue(ret)

    def test_get_object_or_none(self):
        diner = self.loop.run_until_complete(Diner.get_object_or_none(self.db, table_num=1))
        self.assertIsNotNone(diner)
        self.assertEqual(Diner, type(diner))

        # Not existing in the db
        diner = self.loop.run_until_complete(Diner.get_object_or_none(self.db, table_num=999))
        self.assertIsNone(diner)

        diner = self.loop.run_until_complete(Diner.get_object_or_none(self.db, table_num=[1, 2]))
        self.assertIsNotNone(diner)
        self.assertEqual(Diner, type(diner))



import aioredis
import asyncio
from unittest import TestCase


class BaseTestCase(TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        db_co = aioredis.create_redis(
            address=('localhost', 6379),
            db=13,
            loop=self.loop,
            encoding='utf-8',
        )
        self.db = self.loop.run_until_complete(db_co)

    def tearDown(self):
        async def delete_all():
            async for k in self.db.iscan(match='*Test*', count=100):
                await self.db.delete(k)
        self.loop.run_until_complete(delete_all())



from subconscious.model import RedisModel, Column
from uuid import uuid1
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'
    INACTIVE = 'inactive'


class TestUser(RedisModel):
    id = Column(primary_key=True)
    name = Column(index=True)
    age = Column(index=True, type=int)
    locale = Column(index=True, type=int, required=False)
    status = Column(type=str, enum=StatusEnum, index=True)


class TestOrderBy(BaseTestCase):
    def setUp(self):
        super(TestOrderBy, self).setUp()
        self.user_names = []
        for i in range(9):
            user = TestUser(id=str(uuid1()), name='name-{}'.format(i), age=i, locale=i+10, status='active')
            self.loop.run_until_complete(user.save(self.db))
            self.user_names.append(user.name)
        self.user_names_sorted_desc = sorted(self.user_names, reverse=True)

    def test_order_by(self):
        async def _test():
            user_names = [x.name async for x in TestUser.filter_by(self.db, status='active', order_by='name')]
            self.assertEqual(self.user_names, user_names)
            user_names = [x.name async for x in TestUser.filter_by(self.db, status='active', order_by='-name')]
            self.assertEqual(self.user_names_sorted_desc, user_names)

        self.loop.run_until_complete(_test())

    def test_query_with_order_by(self):
        async def _test():
            user_names = [x.name async for x in TestUser.query(self.db).filter(status='active').order_by('name')]
            self.assertEqual(self.user_names, user_names)
            user_names = [x.name async for x in TestUser.query(self.db).filter(status='active').order_by('-name')]
            self.assertEqual(self.user_names_sorted_desc, user_names)
        self.loop.run_until_complete(_test())

    def test_query_order_by_get_first(self):
        async def _test():
            user_name = await TestUser.query(self.db).filter(status='active').order_by('name').first()
            self.assertEqual(self.user_names[0], user_name.name)
            user_name = await TestUser.query(self.db).filter(status='active').order_by('-name').first()
            self.assertEqual(self.user_names[-1], user_name.name)
        self.loop.run_until_complete(_test())

    def test_query_with_limit(self):
        async def _test():
            user_names = [x.name async for x in TestUser.query(self.db).filter(
                status='active').order_by('name').offset(1).limit(3)]
            self.assertEqual(self.user_names[1:4], user_names)
            user_names = [x.name async for x in TestUser.query(self.db).filter(
                status='active').order_by('-name').offset(1).limit(3)]
            self.assertEqual(self.user_names_sorted_desc[1:4], user_names)
        self.loop.run_until_complete(_test())

    def test_empty_filter_by(self):
        async def _test():
            user_names = [x.name async for x in TestUser.query(self.db).order_by('name').offset(1).limit(3)]
            self.assertEqual(self.user_names[1:4], user_names)
        self.loop.run_until_complete(_test())



#!/usr/bin/env python3

import inspect
import logging
import uuid

from .column import Column
from .query import Query


logger = logging.getLogger(__name__)

VALUE_ID_SEPARATOR = '\x00'
MODEL_NAME_ID_SEPARATOR = ':'


# Exceptions

class InvalidQuery(Exception):
    pass


class InvalidModelDefinition(Exception):
    pass


class BadDataError(Exception):
    pass


class UnexpectedColumnError(Exception):
    pass


class ModelMeta(type):

    def __init__(cls, what, bases=None, attributes=None):
        super(ModelMeta, cls).__init__(what, bases, attributes)
        if cls.__name__ not in ('RedisModel', 'TimeStampedModel'):
            columns = []
            num_primary, num_composite = 0, 0
            cls._pk_name = None
            # grab all Columns from the model
            for name, column in inspect.getmembers(cls, lambda col: isinstance(col, Column)):
                column.name = name
                columns.append(column)
                if column.primary:
                    num_primary += 1
                    cls._pk_name = column.name
                if column.composite:
                    num_composite += 1

            # Defensive checks
            if num_primary == 0:
                if num_composite == 0:
                    err_msg = 'No primary key or composite key in {}'.format(cls.__name__)
                    raise InvalidModelDefinition(err_msg)
                if num_composite == 1:
                    err_msg = 'Your composite key is really a primary key in {}'.format(cls.__name__)
                    raise InvalidModelDefinition(err_msg)
            if num_primary == 1:
                if num_composite != 0:
                    err_msg = 'Cannot have both primary and composite keys in {}'.format(cls.__name__)
                    raise InvalidModelDefinition(err_msg)

            cls._columns = tuple(sorted(columns, key=lambda c: c.name))
            cls._indexed_columns = tuple(sorted([col for col in cls._columns if col.indexed], key=lambda c: c.name))
            cls._sortable_columns = tuple(sorted([col for col in cls._columns if col.sorted], key=lambda c: c.name))
            cls._identifier_columns = tuple(
                sorted([col for col in cls._columns if col.primary or col.composite],
                       key=lambda c: c.name))
            cls._auto_columns = sorted(
                [col for col in cls._columns if getattr(col, 'auto_increment', False)],
                key=lambda c: c.name
            )
            cls._queryable_colnames_set = set(
                [col.name for col in cls._indexed_columns + cls._identifier_columns + cls._sortable_columns]
            )
            cls._sortable_column_names = tuple([x.name for x in cls._sortable_columns])
            cls._auto_column_names = {col.name for col in cls._auto_columns}
            cls._indexed_column_names = {col.name for col in cls._indexed_columns}
            cls._columns_map = {c.name: c for c in cls._columns}
            cls._identifier_column_names = tuple([x.name for x in cls._identifier_columns])


class RedisModel(object, metaclass=ModelMeta):

    # force only keyword arguments
    def __init__(self, **kwargs):
        loading = kwargs.pop('loading', False)
        for column in self._columns:
            if column.name in kwargs:
                value = kwargs.pop(column.name)
                if type(value) != column.field_type:
                    err_msg = "Column `{}` in {} has value {}, should be of type {}".format(
                        column.name,
                        self.__class__.__name__,
                        value,
                        column.field_type,
                    )
                    raise BadDataError(err_msg)

                if column.enum_choices and value not in column.enum_choices:
                    err_msg = "Column `{}` in {} has value {}, should be in set {}".format(
                        column.name,
                        self.__class__.__name__,
                        value,
                        column.enum_choices,
                    )
                    raise BadDataError(err_msg)
                if getattr(column, 'auto_increment', False) and not loading:
                    err_msg = "Not allowed to set auto_increment column({})".format(column.name)
                    raise BadDataError(err_msg)

                self.__dict__.update({column.name: value})
            else:
                if column.required and not getattr(column, 'auto_increment', False):
                    err_msg = 'Missing column `{}` in `{}` is required'.format(
                        column.name,
                        self.__class__.__name__,
                    )
                    raise BadDataError(err_msg)

        # Require that every kwarg supplied matches an expected column
        # TODO: handle TimeStampedModel cols better
        known_cols_set = set([column.name for column in self._columns] + ['updated_at', 'created_at'])
        supplied_cols_set = set([x for x in kwargs])
        unknown_cols_set = supplied_cols_set - known_cols_set
        if unknown_cols_set != set():
            err_msg = 'Unknown column(s): {} in `{}`'.format(
                unknown_cols_set,
                self.__class__.__name__,
            )
            raise UnexpectedColumnError(err_msg)

    def __setattr__(self, name, value):
        if name in self._auto_column_names:
            err_msg = "Not allowed to set auto_increment column({})".format(name)
            raise BadDataError(err_msg)

        return super(RedisModel, self).__setattr__(name, value)

    @classmethod
    def key_prefix(cls):
        """Prefix that we use for Redis storage, used for all keys related
        to this object. Default to class name.
        """
        return cls.__name__

    @classmethod
    def make_key(cls, identifier):
        """Convenience method for computing the Redis object instance key
        from the identifier
        """
        return "{}{}{}".format(cls.key_prefix(), MODEL_NAME_ID_SEPARATOR, identifier)

    def has_real_data(self, column_name):
        return not isinstance(getattr(self, column_name), Column)

    def identifier(self):
        identifiers = [str(getattr(self, column.name)) for column in self._identifier_columns]
        return ':'.join(identifiers)

    def redis_key(self):
        """Key used for storage of object instance in Redis.
        """
        return "{}{}{}".format(self.key_prefix(), MODEL_NAME_ID_SEPARATOR, self.identifier())

    def as_dict(self):
        """Dict version of this object
        """
        # WARNING: we have to send a copy, otherwise changing the dict
        # changes the object!
        # FIXME: this returns no keys for keys whose value is None!
        return self.__dict__.copy()

    def __repr__(self):
        return "<{}>".format(self.redis_key())

    @classmethod
    def get_index_key(cls, column_name):
        return 'index{}{}{}{}'.format(MODEL_NAME_ID_SEPARATOR, cls.key_prefix(), MODEL_NAME_ID_SEPARATOR, column_name)

    async def save_index(self, db, stale_object=None):
        for indexed_column in self._queryable_colnames_set:
            index_key = self.get_index_key(indexed_column)
            if stale_object:
                stale_index_value = '{}{}{}'.format(
                    getattr(stale_object, indexed_column),
                    VALUE_ID_SEPARATOR,
                    stale_object.identifier()
                )
                await db.zrem(index_key, stale_index_value)
            index_value = '{}{}{}'.format(
                getattr(self, indexed_column),
                VALUE_ID_SEPARATOR,
                self.identifier()
            )
            # Index it by adding to a sorted set with 0 score. It will be lexically sorted by redis
            await db.zadd(index_key, 0, index_value,)

    async def save(self, db):
        """Save the object to Redis.
        """
        kwargs = {}
        for col in self._auto_columns:
            if not self.has_real_data(col.name):
                kwargs[col.name] = await col.auto_generate(db, self)
        self.__dict__.update(kwargs)

        # we have to delete the old index key
        stale_object = await self.__class__.load(db, identifier=self.identifier())
        success = await db.hmset_dict(self.redis_key(), self.__dict__.copy())
        await self.save_index(db, stale_object=stale_object)
        return success

    async def exists(self, db):
        return await db.exists(self.redis_key())

    @classmethod
    async def load(cls, db, identifier=None, redis_key=None):
        """Load the object from redis. Use the identifier (colon-separated
        composite keys or the primary key) or the redis_key.
        """
        if not identifier and not redis_key:
            raise InvalidQuery('Must supply identifier or redis_key')
        if redis_key is None:
            redis_key = cls.make_key(identifier)
        if await db.exists(redis_key):
            data = await db.hgetall(redis_key)
            kwargs = {}
            for key_bin, value_bin in data.items():
                key, value = key_bin, value_bin
                column = getattr(cls, key, False)
                if not column or (column.field_type == str):
                    kwargs[key] = value
                else:
                    kwargs[key] = column.field_type(value)
            kwargs['loading'] = True
            return cls(**kwargs)
        else:
            logger.debug("No Redis key found: {}".format(redis_key))
            return None

    @classmethod
    async def all(cls, db, order_by=None, limit=None, offset=None):
        async for x in cls.filter_by(db, order_by=order_by, limit=limit, offset=offset):
            yield x

    @classmethod
    async def _get_ordered_result(cls, db, list_to_order, order_by, direction):
        """

        :param list_to_order:
        :param order_by:
        :param direction:
        :return:

        Sort the given list in redis.
        https://redis.io/commands/sort#using-hashes-in-codebycode-and-codegetcode
        """
        pairs = []
        for x in list_to_order:
            pairs.extend([0, x])
        if pairs:
            ordered_res_key = 'filtered_result-{}'.format(uuid.uuid1())
            await db.zadd(ordered_res_key, pairs[0], pairs[1], *pairs[2:])
            ordered_result = await db.sort(
                ordered_res_key,
                by='{}:*->{}'.format(cls.__name__, order_by),
                alpha=True,
                asc=direction
            )
            # Delete the temp store
            await db.delete(ordered_res_key)
            return ordered_result
        else:
            return []

    @classmethod
    async def _get_ids_filter_by(cls, db, order_by=None, **kwargs):
        if order_by:
            direction = b'DESC' if order_by[0] == '-' else None
            if order_by[0] in ('+', '-'):
                order_by = order_by[1:]
                if order_by not in cls._queryable_colnames_set:
                    err_msg = 'order_by field {order_by} is not in {queryable_cols}'.format(
                        order_by=order_by,
                        queryable_cols=cls._queryable_colnames_set,
                    )
                    raise InvalidQuery(err_msg)

        missing_cols_set = set(kwargs.keys()) - cls._queryable_colnames_set
        if missing_cols_set:
            err_msg = '{missing_cols_set} not in {queryable_cols}'.format(
                missing_cols_set=missing_cols_set,
                queryable_cols=cls._queryable_colnames_set,
            )
            raise InvalidQuery(err_msg)
        result_set = set()
        first_iteration = True
        for k, v in kwargs.items():
            if v is None:
                v = cls._columns_map[k]
            if isinstance(v, (list, tuple)):
                values = [str(x) for x in v]
            else:
                values = (str(v),)
            temp_set = set()
            for value in values:
                temp_set = temp_set.union({x.partition(VALUE_ID_SEPARATOR)[2] for x in await db.zrangebylex(
                    cls.get_index_key(k),
                    min='{}{}'.format(value, VALUE_ID_SEPARATOR).encode(),
                    max='{}{}\xff'.format(value, VALUE_ID_SEPARATOR).encode())})
            if first_iteration:
                result_set = result_set.union(temp_set)
                first_iteration = False
            else:
                result_set = result_set.intersection(temp_set)
        if not kwargs:
            for index_entry in await db.zrange(cls.get_index_key(cls._identifier_column_names[0]), 0, -1):
                result_set.add(index_entry.split(VALUE_ID_SEPARATOR)[-1])
        if order_by:
            return await cls._get_ordered_result(db, list_to_order=result_set, order_by=order_by, direction=direction)

        return sorted(result_set)

    @classmethod
    async def filter_by(cls, db, offset=None, limit=None, **kwargs):
        """Query by attributes iteratively. Ordering is not supported
        Example:
            User.get_by(db, age=[32, 54])
            User.get_by(db, age=23, name="guido")

        """
        if limit and type(limit) is not int:
            raise InvalidQuery('If limit is supplied it must be an int')
        if offset and type(offset) is not int:
            raise InvalidQuery('If offset is supplied it must be an int')

        ids_to_iterate = await cls._get_ids_filter_by(db, **kwargs)
        if offset:
            # Using offset without order_by is pretty strange, but allowed
            if limit:
                ids_to_iterate = ids_to_iterate[offset:offset+limit]
            else:
                ids_to_iterate = ids_to_iterate[offset:]
        elif limit:
            ids_to_iterate = ids_to_iterate[:limit]

        for key in ids_to_iterate:
            yield await cls.load(db, key)

    @classmethod
    async def get_object_or_none(cls, db, **kwargs):
        """
        Returns the first object exists for this query or None.
        WARNING: if there are more than 1 results in cls that satisfy the conditions in kwargs,
        only 1 random result will be returned
        """
        async for obj in cls.filter_by(db, limit=1, **kwargs):
            return obj
        return None

    @classmethod
    def query(cls, db) -> Query:
        return Query(model=cls, db=db)



from subconscious.model import RedisModel, Column, InvalidModelDefinition, UnexpectedColumnError
from uuid import uuid1
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'


class TestUser(RedisModel):
    id = Column(primary_key=True)
    name = Column(index=True)
    age = Column(index=True, type=int)
    locale = Column(index=True, type=int, required=False)
    status = Column(type=str, enum=StatusEnum, index=True)


class TestSaveAndLoad(BaseTestCase):

    def test_save_and_load(self):
        user_id = str(uuid1())
        user = TestUser(id=user_id, name='Test name', age=100, status='active')
        ret = self.loop.run_until_complete(user.save(self.db))
        self.assertTrue(ret)

        # load
        user_in_db = self.loop.run_until_complete(TestUser.load(self.db, identifier=user_id))
        self.assertEqual(user_in_db.name, user.name)

    def test_init_model_with_no_indexed_cols_should_error(self):
        with self.assertRaises(InvalidModelDefinition):
            class BadModel(RedisModel):
                unindex_col = Column()


class BadSave(BaseTestCase):

    def test_unexpected_column_should_fail(self):

        class TestModel(RedisModel):
            id = Column(type=int, primary_key=True)

        with self.assertRaises(UnexpectedColumnError):
            TestModel(id=1, this_column_does_not_exist='foo')



#!/usr/bin/env python3

from setuptools import setup, find_packages


setup(
    name='subconscious',
    version='0.08.5',
    packages=find_packages(),
    url='https://github.com/paxos-bankchain/subconscious',
    license='MIT',
    author='Paxos Trust Company, LLC',
    author_email='pypi@paxos.com',
    description='redis-backed db for python3 (asyncio compatible)',
    install_requires=['aioredis'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        # async_generator requires python3.6+
        'Programming Language :: Python :: 3.6',
    ],
)



#!/usr/bin/env python3

from enum import EnumMeta


class InvalidColumnDefinition(Exception):
    pass


class Column(object):
    """Defined fields (columns) for a given RedisModel.
    """

    def __init__(self, type=str, primary_key=None, composite_key=None, index=None,
                 required=None, enum=None, sort=None):
        """primary_key can exist in only a single column.
        composite_key can exist in multiple columns.
        You can't have both a primary_key and composite_key in the same model.
        index is whether you want this column indexed or not for faster retrieval.
        """
        if type not in (str, int):
            # TODO: support for other field types (datetime, uuid, etc)
            err_msg = 'Bad Field Type: {}'.format(type)
            raise InvalidColumnDefinition(err_msg)

        if primary_key and composite_key:
            err_msg = 'Column can be either primary_key or composite_key, but not both'
            raise InvalidColumnDefinition(err_msg)

        self.field_type = type
        self.primary = primary_key is True
        self.composite = composite_key is True
        self.sorted = sort is True
        self.indexed = (index is True) or self.composite
        self.required = required is True or self.primary or self.composite

        self.enum = enum
        if enum:
            if not isinstance(enum, EnumMeta):
                err_msg = '`{}` is not an instance of {}'.format(enum, EnumMeta)
                raise InvalidColumnDefinition(err_msg)
            self.enum_choices = set([x.value for x in enum])
        else:
            self.enum_choices = set()

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)


class Integer(Column):
    def __init__(
            self,
            primary_key=None,
            composite_key=None,
            index=None,
            required=None,
            enum=None,
            sort=None,
            auto_increment=False,):
        super(Integer, self).__init__(
            int,
            primary_key=primary_key,
            composite_key=composite_key,
            index=index,
            required=required,
            enum=enum,
            sort=sort,
        )
        self.auto_increment = auto_increment

    async def auto_generate(self, db, model):
        return await db.incr('auto:{}:{}'.format(model.key_prefix(), self.name))



from subconscious.model import RedisModel, Column
from uuid import uuid1
from .base import BaseTestCase
import enum


class StatusEnum(enum.Enum):
    ACTIVE = 'active'
    INACTIVE = 'inactive'


class TestUser(RedisModel):
    id = Column(primary_key=True)
    name = Column(index=True)
    age = Column(index=True, type=int)
    locale = Column(index=True, type=int, required=False)
    status = Column(type=str, enum=StatusEnum, index=True)


class TestFilterBy(BaseTestCase):
    def setUp(self):
        super(TestFilterBy, self).setUp()
        user = TestUser(id=str(uuid1()), age=0, locale=0+10, status='active')
        self.loop.run_until_complete(user.save(self.db))
        for i in range(9):
            user = TestUser(id=str(uuid1()), name='name-{}'.format(i), age=i, locale=i+10, status='active')
            self.loop.run_until_complete(user.save(self.db))

    def test_filter_by(self):
        async def _test():
            count = 0
            async for x in TestUser.filter_by(self.db, age=1):
                count += 1
            self.assertEqual(1, count)

            count = 0
            async for x in TestUser.filter_by(self.db, age=[1, 2]):
                count += 1
            self.assertEqual(2, count)

            count = 0
            result_list = []
            async for x in TestUser.filter_by(self.db, status='active'):
                count += 1
                self.assertEqual(x.status, 'active')
                result_list.append(x)
            self.assertEqual(10, count)
            result_list[0].status = 'inactive'
            await result_list[0].save(self.db)

            count = 0
            async for x in TestUser.filter_by(self.db, status='active'):
                count += 1
                self.assertEqual(x.status, 'active')
            # Should be one less now
            self.assertEqual(9, count)

        self.loop.run_until_complete(_test())

    def test_get_by_none(self):
        async def _test():
            result_list = []
            async for x in TestUser.filter_by(self.db, name=None):
                result_list.append(x)
            self.assertEqual(1, len(result_list))
        self.loop.run_until_complete(_test())

    def test_query(self):
        async def _test():
            result_list = []
            async for x in TestUser.query(db=self.db).filter(status='active'):
                result_list.append(x)
            self.assertEqual(10, len(result_list))
        self.loop.run_until_complete(_test())

    def test_query_no_filter(self):
        async def _test():
            result_list = []
            async for x in TestUser.query(db=self.db):
                result_list.append(x)
            self.assertEqual(10, len(result_list))
        self.loop.run_until_complete(_test())

    def test_query_first(self):
        async def _test():
            user = await TestUser.query(db=self.db).filter(status='active').first()
            self.assertEqual(TestUser, type(user))
            self.assertEqual(user.status, 'active')
        self.loop.run_until_complete(_test())

    def test_query_first_no_filter(self):
        async def _test():
            user = await TestUser.query(db=self.db).first()
            self.assertEqual(TestUser, type(user))
            self.assertEqual(user.status, 'active')
        self.loop.run_until_complete(_test())

    def test_query_chaining_filters(self):
        async def _test():
            user = await TestUser.query(db=self.db).filter(name='name-1').filter(status='active').first()
            self.assertEqual(TestUser, type(user))
            self.assertEqual(user.status, 'active')
            self.assertEqual(user.name, 'name-1')
        self.loop.run_until_complete(_test())
